import sys
from importlib import import_module
from pathlib import Path

from docker.models.containers import Container


def load_module(config_file: str):
    f = Path(config_file).absolute()
    sys.path.insert(0, str(f.parent))
    return import_module(f.stem)


class BaseConfig:
    @staticmethod
    def create_sender(**kwargs):
        raise NotImplementedError()

    @staticmethod
    def format_log(log_line: str, container: Container):
        return log_line

    @staticmethod
    def filter_log(log_line):
        return True

    @staticmethod
    def filter_container(container: Container) -> bool:
        return True

    BATCH_SIZE = 100
    BATCH_TIMEOUT_SECONDS = 10

    DIE_ON_FAIL = True
    RETRIES = 5


def create_config(config_file: str) -> BaseConfig:
    if not config_file:
        return BaseConfig

    CustomConfig = load_module(config_file).Config

    class Config(CustomConfig, BaseConfig):
        pass

    return Config



import logging
import time
from queue import Empty, Queue
from threading import Thread

import requests

from .config import BaseConfig

logger = logging.getLogger("fourwarder.sender")


class Sender:
    def __init__(self, thread_panic, config: BaseConfig) -> None:
        self.queue = Queue()
        self.terminate_flag = False
        self.thread_panic = thread_panic
        self.config = config

    def start(self):
        self.thread = Thread(target=self.thread_panic(self.run))
        self.thread.start()
        logger.info("started")

    def send_logs(self, logs):
        self.queue.put_nowait(logs)

    def run(self):
        while True:
            try:
                logs = self.queue.get(timeout=0.2)
                self.try_send_data(logs, retries=self.config.RETRIES)
                self.queue.task_done()
            except Empty:
                if self.terminate_flag:
                    return

    def send_data(self, logs):
        raise NotImplementedError()

    def try_send_data(self, logs, backoff=3, retries=5):
        exception = None
        try:
            self.send_data(logs)
        except Exception as e:
            exception = e

        # we run this outside of the try-catch block above to avoid
        # dirty stack traces due to recursion
        if exception:
            logger.error("error when sending data")
            logger.exception(exception)
            if retries > 0:
                logger.info(f"retrying in {backoff} seconds")
                time.sleep(backoff)
                self.try_send_data(logs, backoff * 2, retries - 1)
            else:
                if self.config.DIE_ON_FAIL:
                    raise exception
                logger.warn("max retries exceeded, skipping batch")

    def terminate(self):
        self.terminate_flag = True
        logger.info("terminating")
        self.thread.join()


class HttpSender(Sender):
    def __init__(self, request_kwargs, **kwargs) -> None:
        super().__init__(**kwargs)
        self.request_kwargs = request_kwargs

    def send_data(self, logs):
        requests.request(**self.request_kwargs, json=logs).raise_for_status()



import logging
from threading import RLock, Timer

from .config import BaseConfig
from .sender import Sender

logger = logging.getLogger("fourwarder.batcher")


class Batcher:
    def __init__(self, sender: Sender, thread_panic, config: BaseConfig) -> None:
        self.staged_logs = []
        self.current_batch_timer = None
        self.lock = RLock()
        self.sender = sender
        self.thread_panic = thread_panic
        self.config = config

    def filter_log(self, log_line):
        return self.config.filter_log(log_line)

    def push_line(self, log_line):
        if not self.filter_log(log_line):
            return

        with self.lock:
            self.staged_logs.append(log_line)

            if not self.current_batch_timer:
                self.current_batch_timer = Timer(
                    self.config.BATCH_TIMEOUT_SECONDS, self.thread_panic(self.flush)
                )
                self.current_batch_timer.start()

            if len(self.staged_logs) >= self.config.BATCH_SIZE:
                self.flush()

    def flush(self):
        logger.info("flushing")
        with self.lock:
            if self.staged_logs:
                self.sender.send_logs(self.staged_logs)
                self.staged_logs = []

            if self.current_batch_timer is not None:
                self.current_batch_timer.cancel()
                self.current_batch_timer = None

    def terminate(self):
        self.flush()



import logging
from threading import Thread

from docker.models.containers import Container

from .batcher import Batcher
from .config import BaseConfig


class ContainerHandler:
    def __init__(
        self,
        container: Container,
        batcher: Batcher,
        thread_panic,
        config: BaseConfig,
    ):
        self.container = container
        self.batcher = batcher
        self.logger = logging.getLogger(f"fourwarder.handler.{self.container.name}")
        self.thread_panic = thread_panic
        self.config = config

    def start(self):
        self.log_thread = Thread(target=self.thread_panic(self.watch_logs), daemon=True)
        self.log_stream = self.container.logs(stream=True)
        self.log_thread.start()
        self.logger.info("started")

    def terminate(self):
        self.log_thread.join()
        self.logger.info("terminated")

    def watch_logs(self):
        for line in self.log_stream:
            self.handle_log_line(line)

    def handle_log_line(self, line: str):
        line = line.decode().strip()
        line_data = self.config.format_log(line, self.container)

        self.batcher.push_line(line_data)



import argparse
import logging
import os
import signal
import traceback

import docker
from docker.models.containers import Container

from .batcher import Batcher
from .config import BaseConfig, create_config
from .container_handler import ContainerHandler

logger = logging.getLogger("fourwarder")


class EventHandler:
    def __init__(self, batcher, thread_panic, config: BaseConfig):
        self.client = docker.from_env()
        self.log_handlers = {}
        self.batcher = batcher
        self.own_container_id = os.environ.get("HOSTNAME")
        self.thread_panic = thread_panic
        self.config = config

    def get_event_handler(self, event):
        action = event["Action"]
        type = event["Type"]
        if type != "container":
            return None
        if action == "start":
            return self.handle_start
        if action == "die":
            return self.handle_die

    def handle_event(self, event):
        event_handler = self.get_event_handler(event)
        if not event_handler:
            logger.debug(f'skipping event {event["Type"]} {event["Action"]}')
            return

        event_handler(event)

    def run(self):
        logger.info("starting to read events")
        self.event_stream = self.client.events(decode=True)
        for event in self.event_stream:
            self.handle_event(event)

    def filter_container(self, container: Container):
        if container.attrs["Config"]["Hostname"] == self.own_container_id:
            # make sure we we are not monitoring ourselves
            return False

        return self.config.filter_container(container)

    def handle_start(self, event):
        container = self.client.containers.get(event["Actor"]["ID"])

        if not self.filter_container(container):
            logger.info(f"container {container.name} excluded")
            return

        log_handler = ContainerHandler(
            container, self.batcher, self.thread_panic, self.config
        )
        self.log_handlers[container.id] = log_handler
        log_handler.start()

    def handle_die(self, event):
        container_id = event["Actor"]["ID"]
        if container_id not in self.log_handlers:
            return

        self.log_handlers[container_id].terminate()
        del self.log_handlers[container_id]

    def terminate(self):
        self.event_stream.close()


class Main:
    def __init__(self, config: BaseConfig):
        self.sender = config.create_sender(
            thread_panic=self.thread_panic, config=config
        )
        self.batcher = Batcher(self.sender, self.thread_panic, config)
        self.event_handler = EventHandler(self.batcher, self.thread_panic, config)
        self.termination_exc = None

    def thread_panic(self, func):
        """wraps the passed function into a try catch that terminates the main thread."""

        def wrapper():
            try:
                func()
            except Exception:
                self.event_handler.terminate()
                self.termination_exc = traceback.format_exc()

        return wrapper

    def run(self):
        self.sender.start()
        self.event_handler.run()

    def terminate(self):
        self.batcher.terminate()
        self.sender.terminate()
        self.event_handler.terminate()


def run():
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument("config_path")
    config_path = parser.parse_args().config_path
    config = create_config(config_path)

    main = Main(config)

    terminate_signal_sent = False

    def terminate_signal_handler(x, y):
        logger.warn("received termination signal")
        nonlocal terminate_signal_sent
        if not terminate_signal_sent:
            terminate_signal_sent = True
            main.terminate()
        else:
            # exit if pressed twice
            exit(1)

    signal.signal(signal.SIGINT, terminate_signal_handler)

    main.run()

    if main.termination_exc:
        logger.error(main.termination_exc)
        exit(1)
